import numpy as np

from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    IMAGEDATA_ELEMENT,
    NODE_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    Dims,
    Encoding,
    ImageData,
)

from .base_addon import BaseCorrectionAddon


def maybe_get(a):
    # TODO: proper check for cupy
    if hasattr(a, "get"):
        return a.get()
    return a


class SaturationMonitor(BaseCorrectionAddon):
    def __init__(self, config):
        global cupy
        import cupy
        self._alarmThreshold = config["alarmThreshold"]
        self._warnThreshold = config["warnThreshold"]
        self._alarmMaxCount = config["alarmMaxCount"]
        self._warnMaxCount = config["warnMaxCount"]
        self._frameAxis = config["frameAxis"]

    def reconfigure(self, changed_config):
        if changed_config.has("alarmThreshold"):
            self._alarmThreshold = changed_config["alarmThreshold"]
        if changed_config.has("warnThreshold"):
            self._warnThreshold = changed_config["warnThreshold"]
        if changed_config.has("alarmMaxCount"):
            self._alarmMaxCount = changed_config["alarmMaxCount"]
        if changed_config.has("warnMaxCount"):
            self._warnMaxCount = changed_config["warnMaxCount"]
        if changed_config.has("frameAxis"):
            self._frameAxis = changed_config["frameAxis"]

    @staticmethod
    def extend_output_schema(schema):
        (
            NODE_ELEMENT(schema)
            .key("saturationMonitor")
            .commit(),

            BOOL_ELEMENT(schema)
            .key("saturationMonitor.warning")
            .readOnly()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("saturationMonitor.alarm")
            .readOnly()
            .commit(),

            UINT32_ELEMENT(schema)
            .key("saturationMonitor.warnCount")
            .description(
                "Total number of pixels above warning threshold. Each pixel "
                "is only counted once even if it exceeds the threshold in "
                "multiple frames (/ memory cells = given axis)."
            )
            .readOnly()
            .commit(),

            UINT32_ELEMENT(schema)
            .key("saturationMonitor.alarmCount")
                .description(
                "Total number of pixels above alarm threshold. Each pixel "
                "is only counted once even if it exceeds the threshold in "
                "multiple frames (/ memory cells = given axis)."
            )
            .readOnly()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("saturationMonitor.maxValue")
            .readOnly()
            .commit(),

            # TODO: switch to image data
            IMAGEDATA_ELEMENT(schema)
            .key("saturationMonitor.alarmImage")
            .commit(),
        )

    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.alarmThreshold")
            .description("Alarm threshold per pixel.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.warnThreshold")
            .description("Warning threshold per pixel.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(schema)
            .key(f"{prefix}.alarmMaxCount")
            .description("Maximum number of pixel above alarm threshold.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(schema)
            .key(f"{prefix}.warnMaxCount")
            .description("Maximum number of pixel above warn threshold.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(schema)
            .key(f"{prefix}.frameAxis")
            .displayedName('Multi-frame axis')
            .description("Axis for frames. Used to take the max over this axis.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
        )

    def post_correction(self, processed_data, cell_table, pulse_table, output_hash):
        # only take the max if data has frames -> more than 2 dimensions
        if processed_data.ndim > 2:
            max_image = np.nanmax(processed_data, axis=self._frameAxis)
        else:
            max_image = processed_data
        nb_pix_warning = int(np.nansum(max_image > self._warnThreshold))
        nb_pix_alarm = int(np.nansum(max_image > self._alarmThreshold))

        output_hash["saturationMonitor.warning"] = nb_pix_warning > self._warnMaxCount
        output_hash["saturationMonitor.alarm"] = nb_pix_alarm > self._alarmMaxCount
        output_hash["saturationMonitor.warnCount"] = nb_pix_warning
        output_hash["saturationMonitor.alarmCount"] = nb_pix_alarm
        output_hash["saturationMonitor.maxValue"] = float(np.nanmax(max_image))
        max_image[max_image <= self._warnThreshold] = 0
        output_hash["saturationMonitor.alarmImage"] = ImageData(
            maybe_get(max_image), Dims(*max_image.shape), Encoding.GRAY, bitsPerPixel=32
        )
