import enum

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    FLOAT_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,

    schemas,
    utils,
)
from ..preview_utils import FrameSelectionMode, PreviewFriend, PreviewSpec
from .._version import version as deviceVersion


class Constants(enum.Enum):
    ThresholdsDark = enum.auto()
    Offset = enum.auto()
    SlopesCS = enum.auto()
    SlopesPC = enum.auto()
    SlopesFF = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsCS = enum.auto()
    BadPixelsPC = enum.auto()
    BadPixelsFF = enum.auto()


bad_pixel_constants = {
    Constants.BadPixelsDark,
    Constants.BadPixelsCS,
    Constants.BadPixelsPC,
    Constants.BadPixelsFF,
}


# from pycalibration's enum.py
class GainModes(enum.IntEnum):
    ADAPTIVE_GAIN = 0
    FIXED_HIGH_GAIN = 1
    FIXED_MEDIUM_GAIN = 2
    FIXED_LOW_GAIN = 3


# note: if this is extended further, bump up the dtype passed to kernel
class CorrectionFlags(enum.IntFlag):
    NONE = 0
    THRESHOLD = 1
    OFFSET = 2
    REL_GAIN_PC = 4
    GAIN_XRAY = 8
    BPMASK = 16
    FORCE_MG_IF_BELOW = 32
    FORCE_HG_IF_BELOW = 64
    COMMON_MODE = 128


correction_steps = (
    # step name (used in schema), flag to enable for kernel, constants required
    ("thresholding", CorrectionFlags.THRESHOLD, {Constants.ThresholdsDark}),
    ("offset", CorrectionFlags.OFFSET, {Constants.Offset}),
    (  # TODO: specify that /both/ constants are needed, not just one or the other
        "forceMgIfBelow",
        CorrectionFlags.FORCE_MG_IF_BELOW,
        {Constants.ThresholdsDark, Constants.Offset},
    ),
    (
        "forceHgIfBelow",
        CorrectionFlags.FORCE_HG_IF_BELOW,
        {Constants.ThresholdsDark, Constants.Offset},
    ),
    ("relGain", CorrectionFlags.REL_GAIN_PC, {Constants.SlopesCS, Constants.SlopesPC}),
    ("gainXray", CorrectionFlags.GAIN_XRAY, {Constants.SlopesFF}),
    (
        "badPixels",
        CorrectionFlags.BPMASK,
        bad_pixel_constants | {
            None,  # means stay available even without constants loaded
        },
    ),
    ("commonMode", CorrectionFlags.COMMON_MODE, {None}),
)


class AgipdBaseRunner(base_kernel_runner.BaseKernelRunner):
    _bad_pixel_constants = bad_pixel_constants
    _correction_flag_class = CorrectionFlags
    _correction_steps = correction_steps
    num_pixels_ss = 512
    num_pixels_fs = 128

    @classmethod
    def add_schema(cls, schema):
        # will add node, enable / preview toggles
        super(cls, cls).add_schema(schema)
        super(cls, cls).add_bad_pixel_config(schema)

        # additional settings specific to AGIPD correction steps
        # turn off some advanced corrections by default
        for flag in ("enable", "preview"):
            (
                OVERWRITE_ELEMENT(schema)
                .key(f"corrections.commonMode.{flag}")
                .setNewDefaultValue(False)
                .commit(),
            )
            for step in ("forceMgIfBelow", "forceHgIfBelow"):
                (
                    OVERWRITE_ELEMENT(schema)
                    .key(f"corrections.{step}.{flag}")
                    .setNewDefaultValue(False)
                    .commit(),
                )
        (
            FLOAT_ELEMENT(schema)
            .key("corrections.forceMgIfBelow.hardThreshold")
            .tags("managed")
            .description(
                "If enabled, any pixels assigned to low gain stage which would be "
                "below this threshold after having medium gain offset subtracted will "
                "be reassigned to medium gain."
            )
            .assignmentOptional()
            .defaultValue(1000)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("corrections.forceHgIfBelow.hardThreshold")
            .tags("managed")
            .description(
                "Like forceMgIfBelow, but potentially reassigning from medium gain to "
                "high gain based on threshold and pixel value minus low gain offset. "
                "Applied after forceMgIfBelow, pixels can theoretically be reassigned "
                "twice, from LG to MG and to HG."
            )
            .assignmentOptional()
            .defaultValue(1000)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("corrections.relGain.sourceOfSlopes")
            .tags("managed")
            .displayedName("Kind of slopes")
            .description(
                "Slopes for relative gain correction can be derived from pulse "
                "capacitor (PC) scans or current source scans (CS). Choose one!"
            )
            .assignmentOptional()
            .defaultValue("PC")
            .options("PC,CS")
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("corrections.relGain.adjustMgBaseline")
            .tags("managed")
            .displayedName("Adjust MG baseline")
            .description(
                "If set, an additional offset is applied to pixels in medium gain. "
                "The value of the offset is either computed per pixel (and memory "
                "cell) based on currently loaded relative gain slopes (PC or CS) or "
                "it is set to a specific value when overrideMdAdditionalOffset is set."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("corrections.relGain.overrideMdAdditionalOffset")
            .tags("managed")
            .displayedName("Override md_additional_offset")
            .description(
                "If set, the additional offset applied to medium gain pixels "
                "(assuming adjustMgBaseLine is set) will use the value from "
                "mdAdditionalOffset globally for. Otherwise, the additional offset "
                "is derived from the relevant constants."
            )
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("corrections.relGain.mdAdditionalOffset")
            .tags("managed")
            .displayedName("Value for md_additional_offset (if overriding)")
            .description(
                "Normally, md_additional_offset (part of relative gain correction) is "
                "computed when loading SlopesPC. In case you want to use a different "
                "value (global for all medium gain pixels), you can specify it here "
                "and set corrections.overrideMdAdditionalOffset to True."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("corrections.gainXray.gGainValue")
            .tags("managed")
            .displayedName("G_gain_value")
            .description(
                "Newer X-ray gain correction constants are absolute. The default "
                "G_gain_value of 1 means that output is expected to be in keV. If "
                "this is not desired, one can here specify the mean X-ray gain value "
                "over all modules to get ADU values out - operator must manually "
                "find this mean value."
            )
            .assignmentOptional()
            .defaultValue(1)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key("corrections.commonMode.iterations")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(4)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("corrections.commonMode.minFrac")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0.15)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("corrections.commonMode.noisePeakRange")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(35)
            .reconfigurable()
            .commit(),
        )

    def reconfigure(self, config):
        super().reconfigure(config)

        if config.has("corrections.gainXray.gGainValue"):
            self.g_gain_value = self._xp.float32(
                config["corrections.gainXray.gGainValue"]
            )

        if config.has("corrections.forceHgIfBelow.hardThreshold"):
            self.hg_hard_threshold = self._xp.float32(
                config["corrections.forceHgIfBelow.hardThreshold"]
            )

        if config.has("corrections.forceMgIfBelow.hardThreshold"):
            self.mg_hard_threshold = self._xp.float32(
                config["corrections.forceMgIfBelow.hardThreshold"]
            )

        if (
            config.has("corrections.relGain.overrideMdAdditionalOffset")
            or config.has("corrections.relGain.mdAdditionalOffset")
        ):
            if (
                self._get("corrections.relGain.overrideMdAdditionalOffset")
                and self._get("corrections.relGain.adjustMgBaseLine")
            ):
                self.md_additional_offset.fill(
                    self._xp.float32(
                        self._get("corrections.relGain.mdAdditionalOffset")
                    )
                )
            else:
                self._device._please_send_me_cached_constants(
                    {
                        Constants.SlopesPC
                        if self._get("corrections.relGain.sourceOfSlopes") == "PC"
                        else Constants.SlopesCS
                    },
                    self.load_constant,
                )

        if config.has("corrections.relGain.sourceOfSlopes"):
            self.flush_buffers(
                {
                    Constants.SlopesPC,
                    Constants.SlopesCS,
                    Constants.BadPixelsCS,
                    Constants.BadPixelsPC,
                }
            )

            # note: will try to reload all bad pixels, but
            # _load_constant will reject PC/CS based on which we want
            if self._get("corrections.relGain.sourceOfSlopes") == "PC":
                to_get = {Constants.SlopesPC} | bad_pixel_constants
            else:
                to_get = {Constants.SlopesCS} | bad_pixel_constants

            self._device._please_send_me_cached_constants(to_get, self.load_constant)

        if config.has("corrections.commonMode"):
            self.cm_min_frac = self._xp.float32(
                self._get("corrections.commonMode.minFrac")
            )
            self.cm_noise_peak = self._xp.float32(
                self._get("corrections.commonMode.noisePeakRange")
            )
            self.cm_iter = self._xp.uint16(
                self._get("corrections.commonMode.iterations")
            )

        if config.has("constantParameters.gainMode"):
            gain_mode = GainModes[config["constantParameters.gainMode"]]
            if gain_mode is GainModes.ADAPTIVE_GAIN:
                self.default_gain = self._xp.uint8(gain_mode)
            else:
                self.default_gain = self._xp.uint8(gain_mode - 1)

    def expected_input_shape(self, num_frames):
        return (
            num_frames,
            2,
            self.num_pixels_ss,
            self.num_pixels_fs,
        )

    @property
    def _gm_map_shape(self):
        return self._map_shape + (3,)  # for gain-mapped constants

    @property
    def threshold_map_shape(self):
        return self._map_shape + (2,)

    def _setup_constant_buffers(self):
        # constants
        self.gain_thresholds = self._xp.empty(
            self.threshold_map_shape, dtype=np.float32
        )
        self.offset_map = self._xp.empty(self._gm_map_shape, dtype=np.float32)
        self.rel_gain_map = self._xp.empty(self._gm_map_shape, dtype=np.float32)
        # not _gm_map_shape because it only applies to medium gain pixels
        self.md_additional_offset = self._xp.empty(self._map_shape, dtype=np.float32)
        self.xray_gain_map = self._xp.empty(self._map_shape, dtype=np.float32)
        self.bad_pixel_map = self._xp.empty(self._gm_map_shape, dtype=np.uint32)
        self.flush_buffers(set(Constants))

    def _make_output_buffers(self, num_frames, flags):
        output_shape = (num_frames, self.num_pixels_ss, self.num_pixels_fs)
        return [
            self._xp.empty(output_shape, dtype=self._xp.float32),  # image
            self._xp.empty(output_shape, dtype=self._xp.float32),  # gain
        ]

    def _preview_data_views(self, raw_data, processed_data, gain_map):
        return (
            raw_data[:, 0],  # raw
            processed_data,  # corrected
            raw_data[:, 1],  # raw gain
            gain_map,  # digitized gain
        )

    def _load_constant(self, constant, constant_data):
        if constant is Constants.ThresholdsDark:
            # shape: y, x, memory cell, thresholds and gain values
            # note: the gain values are something like means used to derive thresholds
            self.gain_thresholds[:] = self._xp.asarray(
                constant_data[..., :2], dtype=np.float32
            ).transpose((2, 1, 0, 3))[:self._constant_memory_cells]
        elif constant is Constants.Offset:
            # shape: y, x, memory cell, gain stage
            self.offset_map[:] = self._xp.asarray(
                constant_data, dtype=np.float32
            ).transpose((2, 1, 0, 3))[:self._constant_memory_cells]
        elif constant is Constants.SlopesCS:
            if self._get("corrections.relGain.sourceOfSlopes") == "PC":
                return
            self.rel_gain_map.fill(1)
            self.rel_gain_map[..., 1] = self.rel_gain_map[..., 0] * self._xp.asarray(
                constant_data[ ..., :self._constant_memory_cells, 6]
            ).T
            self.rel_gain_map[..., 2] = self.rel_gain_map[..., 1] * self._xp.asarray(
                constant_data[ ..., :self._constant_memory_cells, 7]
            ).T
            if self._get("corrections.relGain.adjustMgBaseline"):
                self.md_additional_offset.fill(
                    self._get("corrections.relGain.mdAdditionalOffset")
                )
            else:
                self.md_additional_offset.fill(0)
        elif constant is Constants.SlopesPC:
            if self._get("corrections.relGain.sourceOfSlopes") == "CS":
                return
            # pc has funny shape (11, 352, 128, 512) from file (fi, memory cell, y, x)
            # assume newer variant (TODO: check) which is pre-sanitized
            hg_slope = constant_data[0]
            hg_intercept = constant_data[1]
            mg_slope = constant_data[3]
            mg_intercept = constant_data[4]

            hg_slope_median = np.nanmedian(hg_slope, axis=(1, 2))
            mg_slope_median = np.nanmedian(mg_slope, axis=(1, 2))

            frac_hg_mg = hg_slope / mg_slope
            rel_gain_map = np.ones(
                (
                    3,
                    self._constant_memory_cells,
                    self.num_pixels_fs,
                    self.num_pixels_ss,
                ),
                dtype=np.float32,
            )
            rel_gain_map[1] = rel_gain_map[0] * frac_hg_mg
            rel_gain_map[2] = rel_gain_map[1] * 4.48
            self.rel_gain_map[:] = self._xp.asarray(
                rel_gain_map.transpose((1, 3, 2, 0)), dtype=np.float32
            )[:self._constant_memory_cells]
            if self._get("corrections.relGain.adjustMgBaseline"):
                if self._get("corrections.relGain.overrideMdAdditionalOffset"):
                    self.md_additional_offset.fill(
                        self._get("corrections.relGain.mdAdditionalOffset")
                    )
                else:
                    self.md_additional_offset[:] = self._xp.asarray(
                        (
                            hg_intercept - mg_intercept * frac_hg_mg
                        ).astype(np.float32).transpose((0, 2, 1)), dtype=np.float32
                    )[:self._constant_memory_cells]
            else:
                self.md_additional_offset.fill(0)
        elif constant is Constants.SlopesFF:
            # constant shape: y, x, memory cell
            if constant_data.shape[2] == 2:
                # TODO: remove support for old format
                # old format, is per pixel only (shape is y, x, 2)
                # note: we should not support this in online
                constant_data = self._xp.broadcast_to(
                    constant_data[..., 0][..., np.newaxis],
                    (
                        self.num_pixels_fs,
                        self.num_pixels_ss,
                        self._constant_memory_cells,
                    ),
                )
            self.xray_gain_map[:] = self._xp.asarray(
                constant_data.transpose(), dtype=np.float32
            )[:self._constant_memory_cells]
        else:
            assert constant in bad_pixel_constants
            if (
                constant is Constants.BadPixelsCS
                and self._get("corrections.relGain.sourceOfSlopes") == "PC"
                or constant is Constants.BadPixelsPC
                and self._get("corrections.relGain.sourceOfSlopes") == "CS"
            ):
                return
            # will simply OR with already loaded, does not take into account which ones
            constant_data = self._xp.asarray(constant_data, dtype=np.uint32)
            if len(constant_data.shape) == 3:
                if constant_data.shape[:2] == (
                    self.num_pixels_fs,
                    self.num_pixels_ss,
                ):
                    # BadPixelsFF is not per gain stage - broadcasting along gain
                    constant_data = self._xp.broadcast_to(
                        constant_data.transpose()[
                            :self._constant_memory_cells, ..., np.newaxis
                        ],
                        self._gm_map_shape,
                    )
                elif constant_data.shape[1:] == (
                    self.num_pixels_fs,
                    self.num_pixels_ss,
                ):
                    # old BadPixelsPC have different axis order
                    constant_data = self._xp.broadcast_to(
                        constant_data.transpose((0, 2, 1))[
                            :self._constant_memory_cells, ..., np.newaxis
                        ],
                        self._gm_map_shape,
                    )
                else:
                    raise ValueError(
                        f"Unsupported bad pixel map shape: {constant_data.shape}"
                    )
            else:
                # gain mapped constants seem consistent
                constant_data = constant_data.transpose((2, 1, 0, 3))
            constant_data &= self._xp.uint32(self.bad_pixel_subset)
            self.bad_pixel_map |= constant_data[:self._constant_memory_cells]

    def flush_buffers(self, constants):
        if Constants.Offset in constants:
            self.offset_map.fill(0)
        if constants & {Constants.SlopesCS, Constants.SlopesPC}:
            self.rel_gain_map.fill(1)
            self.md_additional_offset.fill(0)
        if Constants.SlopesFF:
            self.xray_gain_map.fill(1)
        if constants & bad_pixel_constants:
            self.bad_pixel_map.fill(0)
            if self.bad_pixel_subset & utils.BadPixelValues.NON_STANDARD_SIZE:
                self._mask_asic_seams

    def _mask_asic_seams(self):
        self.bad_pixel_map[:, 64:512:64] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
        self.bad_pixel_map[:, 63:511:64] |= utils.BadPixelValues.NON_STANDARD_SIZE.value


class AgipdCpuRunner(AgipdBaseRunner):
    _xp = np

    def _correct(self, flags, image_data, cell_table, processed_data, gain_map):
        if flags & CorrectionFlags.COMMON_MODE:
            raise NotImplementedError("Common mode not available on CPU yet")
        self.correction_kernel(
            image_data,
            cell_table,
            flags,
            self.default_gain,
            self.gain_thresholds,
            self.mg_hard_threshold,
            self.hg_hard_threshold,
            self.offset_map,
            self.rel_gain_map,
            self.md_additional_offset,
            self.xray_gain_map,
            self.g_gain_value,
            self.bad_pixel_map,
            self.bad_pixel_mask_value,
            processed_data,
        )

    def _pre_init(self):
        from ..kernels import agipd_cython
        self.correction_kernel = agipd_cython.correct


class AgipdGpuRunner(AgipdBaseRunner):
    def _pre_init(self):
        import cupy
        self._xp = cupy

    def _post_init(self):
        self.correction_kernel = self._xp.RawKernel(
            code=base_kernel_runner.get_kernel_template("agipd_gpu.cu").render(
                pixels_x=self.num_pixels_ss,
                pixels_y=self.num_pixels_fs,
                output_data_dtype=utils.np_dtype_to_c_type(self._output_dtype),
                corr_enum=utils.enum_to_c_template(CorrectionFlags),
            ),
            name="correct",
        )
        self.cm_kernel = self._xp.RawKernel(
            code=base_kernel_runner.get_kernel_template("common_gpu.cu").render(
                ss_dim=self.num_pixels_ss,
                fs_dim=self.num_pixels_fs,
                asic_dim=64,
            ),
            name="common_mode_asic",
        )

    def _correct(self, flags, image_data, cell_table, processed_data, gain_map):
        num_frames = self._xp.uint16(image_data.shape[0])
        block_shape = (1, 1, 64)
        grid_shape = utils.grid_to_cover_shape_with_blocks(
            processed_data.shape, block_shape
        )

        self.correction_kernel(
            grid_shape,
            block_shape,
            (
                image_data,
                cell_table,
                flags,
                num_frames,
                self._constant_memory_cells,
                self.default_gain,
                self.gain_thresholds,
                self.mg_hard_threshold,
                self.hg_hard_threshold,
                self.offset_map,
                self.rel_gain_map,
                self.md_additional_offset,
                self.xray_gain_map,
                self.g_gain_value,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                gain_map,
                processed_data,
            ),
        )

        if flags & CorrectionFlags.COMMON_MODE:
            # grid: one block per ss asic, fs asic, and frame
            self.cm_kernel(
                (2, 8, processed_data.shape[0]),
                (64, 1, 1),
                (
                    processed_data,
                    num_frames,
                    self.cm_iter,
                    self.cm_min_frac,
                    self.cm_noise_peak,
                ),
            )


class AgipdCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.ThresholdsDark: self.dark_condition,
            Constants.Offset: self.dark_condition,
            Constants.SlopesCS: self.dark_condition,
            Constants.SlopesPC: self.dark_condition,
            Constants.SlopesFF: self.illuminated_condition,
            Constants.BadPixelsDark: self.dark_condition,
            Constants.BadPixelsCS: self.dark_condition,
            Constants.BadPixelsPC: self.dark_condition,
            Constants.BadPixelsFF: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(AgipdCalcatFriend, AgipdCalcatFriend).add_schema(schema, "AGIPD-Type")

        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(352)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(300)
            .commit()
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.acquisitionRate")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.gainSetting")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.photonEnergy")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(9.2)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.gainMode")
            .tags("managed")
            .assignmentOptional()
            .defaultValue("ADAPTIVE_GAIN")
            .options(",".join(gain_mode.name for gain_mode in GainModes))
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(12)
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Acquisition rate"] = self._get_param("acquisitionRate")

        # TODO: remove this workaround after CalCat update
        integration_time = self._get_param("integrationTime")
        if integration_time != 12:
            res["Integration Time"] = integration_time

        gain_mode = GainModes[self._get_param("gainMode")]
        if gain_mode is not GainModes.ADAPTIVE_GAIN:
            res["Gain Mode"] = 1

        # TODO: make configurable whether or not to include gain setting?
        res["Gain Setting"] = self._get_param("gainSetting")

        return res

    def illuminated_condition(self):
        res = self.dark_condition()

        # note: can consider always setting memory cells to 352 for FF
        # (deviation on constants in database should remove need for this, though)

        # for now, FF constants don't care about gain mode
        if "Gain Mode" in res:
            del res["Gain Mode"]

        res["Source Energy"] = self._get_param("photonEnergy")

        return res


@KARABO_CLASSINFO("AgipdCorrection", deviceVersion)
class AgipdCorrection(base_correction.BaseCorrection):
    # subclass *must* set these attributes
    _base_output_schema = schemas.xtdf_output_schema
    _calcat_friend_class = AgipdCalcatFriend
    _constant_enum_class = Constants
    _correction_steps = correction_steps

    _preview_outputs = [
        PreviewSpec(name, frame_reduction_selection_mode=FrameSelectionMode.CELL)
        for name in ("raw", "corrected", "rawGain", "gainMap")
    ]

    @classmethod
    def expectedParameters(cls, expected):
        # this is not automatically done by superclass for complicated class reasons
        cls._calcat_friend_class.add_schema(expected)
        AgipdBaseRunner.add_schema(expected)
        base_correction.add_addon_nodes(expected, cls)
        PreviewFriend.add_schema(expected, cls._preview_outputs)
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(
                cls._base_output_schema(use_shmem_handles=cls._use_shmem_handles)
            )
            .commit(),
        )

        (
            # support both CPU and GPU kernels
            STRING_ELEMENT(expected)
            .key("kernelType")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(base_kernel_runner.KernelRunnerTypes.GPU.name)
            .options(
                ",".join(
                    kernel_type.name
                    for kernel_type in base_kernel_runner.KernelRunnerTypes
                )
            )
            .reconfigurable()
            .commit(),
        )

    @property
    def _kernel_runner_class(self):
        kernel_type = base_kernel_runner.KernelRunnerTypes[
            self.unsafe_get("kernelType")
        ]
        if kernel_type is base_kernel_runner.KernelRunnerTypes.CPU:
            return AgipdCpuRunner
        else:
            return AgipdGpuRunner
