#include <cuda_fp16.h>

{{corr_enum}}

extern "C" {
	/*
	  Perform corrections; see agipd_gpu.CorrectionFlags
	  Note that THRESHOLD and OFFSET should for any later corrections to make sense
	  Will take cell_table into account when getting correction values
	  Will convert from input dtype to float for correction
	  Will convert to output dtype for output
	*/
	__global__ void correct(const unsigned short* data,
	                        const unsigned short* cell_table,
	                        const unsigned char corr_flags,
	                        const unsigned short input_frames,
	                        const unsigned short map_cells,
	                        // default_gain can be 0, 1, or 2, and is relevant for fixed gain mode (no THRESHOLD)
	                        const unsigned char default_gain,
	                        const float* threshold_map,
	                        const float mg_hard_threshold,
	                        const float hg_hard_threshold,
	                        const float* offset_map,
	                        const float* rel_gain_pc_map,
	                        const float* md_additional_offset,
	                        const float* rel_gain_xray_map,
	                        const float g_gain_value,
	                        const unsigned int* bad_pixel_map,
	                        const float bad_pixel_mask_value,
	                        float* gain_map, // TODO: more compact yet plottable representation
	                        {{output_data_dtype}}* output) {
		const size_t ss_dim = 512;
		const size_t fs_dim = 128;

		const size_t frame = blockIdx.x * blockDim.x + threadIdx.x;
		const size_t ss = blockIdx.y * blockDim.y + threadIdx.y;
		const size_t fs = blockIdx.z * blockDim.z + threadIdx.z;

		if (frame >= input_frames || fs >= fs_dim || ss >= ss_dim) {
			return;
		}

		// data shape: frame, data/raw_gain (dim size 2), ss, fs
		const size_t data_stride_fs = 1;
		const size_t data_stride_ss = fs_dim * data_stride_fs;
		const size_t data_stride_raw_gain = ss_dim * data_stride_ss;
		const size_t data_stride_frame = 2 * data_stride_raw_gain;
		const size_t data_index = frame * data_stride_frame +
			0 * data_stride_raw_gain +
			fs * data_stride_fs +
			ss * data_stride_ss;
		const size_t raw_gain_index = frame * data_stride_frame +
			1 * data_stride_raw_gain +
			fs * data_stride_fs +
			ss * data_stride_ss;
		float res = (float)data[data_index];
		const float raw_gain_val = (float)data[raw_gain_index];

		const size_t output_stride_fs = 1;
		const size_t output_stride_ss = output_stride_fs * fs_dim;
		const size_t output_stride_frame = output_stride_ss * ss_dim;
		const size_t output_index = frame * output_stride_frame + ss * output_stride_ss + fs * output_stride_fs;

		// per-pixel only constant: cell, ss, fs
		const size_t map_stride_fs = 1;
		const size_t map_stride_ss = fs_dim * map_stride_fs;
		const size_t map_stride_cell = ss_dim * map_stride_ss;

		// threshold constant shape: cell, ss, fs, threshold (dim size 2)
		const size_t threshold_map_stride_threshold = 1;
		const size_t threshold_map_stride_fs = 2 * threshold_map_stride_threshold;
		const size_t threshold_map_stride_ss = fs_dim * threshold_map_stride_fs;
		const size_t threshold_map_stride_cell = ss_dim * threshold_map_stride_ss;

		// gain mapped constant shape: cell, ss, fs, gain_level (dim size 3)
		const size_t gm_map_stride_gain = 1;
		const size_t gm_map_stride_fs = 3 * gm_map_stride_gain;
		const size_t gm_map_stride_ss = fs_dim * gm_map_stride_fs;
		const size_t gm_map_stride_cell = ss_dim * gm_map_stride_ss;
		// note: assuming all maps have same shape (in terms of cells / ss / fs)

		const size_t map_cell = cell_table[frame];

		if (map_cell < map_cells) {
			unsigned char gain = default_gain;
			if (corr_flags & THRESHOLD) {
				const float threshold_0 = threshold_map[0 * threshold_map_stride_threshold +
				                                        map_cell * threshold_map_stride_cell +
				                                        fs * threshold_map_stride_fs +
				                                        ss * threshold_map_stride_ss];
				const float threshold_1 = threshold_map[1 * threshold_map_stride_threshold +
				                                        map_cell * threshold_map_stride_cell +
				                                        fs * threshold_map_stride_fs +
				                                        ss * threshold_map_stride_ss];
				// could consider making this const using ternaries / tiny function
				if (raw_gain_val <= threshold_0) {
					gain = 0;
				} else if (raw_gain_val <= threshold_1) {
					gain = 1;
				} else {
					gain = 2;
				}
			}

			const size_t gm_map_index_without_gain = map_cell * gm_map_stride_cell +
				fs * gm_map_stride_fs +
				ss * gm_map_stride_ss;
			if ((corr_flags & FORCE_MG_IF_BELOW) && (gain == 2) &&
			    (res - offset_map[gm_map_index_without_gain + 1 * gm_map_stride_gain] < mg_hard_threshold)) {
				gain = 1;
			}
			if ((corr_flags & FORCE_HG_IF_BELOW) && (gain == 1) &&
			    (res - offset_map[gm_map_index_without_gain] < hg_hard_threshold)) {
				gain = 0;
			}
			gain_map[output_index] = (float)gain;

			const size_t map_index = map_cell * map_stride_cell +
				fs * map_stride_fs +
				ss * map_stride_ss;

			const size_t gm_map_index = gm_map_index_without_gain + gain * gm_map_stride_gain;

			if ((corr_flags & BPMASK) && bad_pixel_map[gm_map_index]) {
				res = bad_pixel_mask_value;
				gain_map[output_index] = bad_pixel_mask_value;
			} else {
				if (corr_flags & OFFSET) {
					res -= offset_map[gm_map_index];
					// TODO: optionally reassign gain stage for this pixel based on new value
				}
				// TODO: baseline shift
				if (corr_flags & REL_GAIN_PC) {
					res *= rel_gain_pc_map[gm_map_index];
					if (gain == 1) {
						res += md_additional_offset[map_index];
					}
				}
				if (corr_flags & GAIN_XRAY) {
					res = (res / rel_gain_xray_map[map_index]) * g_gain_value;
				}
			}
		}
		{% if output_data_dtype == "half" %}
		output[output_index] = __float2half(res);
		{% else %}
		output[output_index] = ({{output_data_dtype}})res;
		{% endif %}
	}
}
