import enum
import functools
from dataclasses import dataclass
from typing import Callable, List, Optional

from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    INT32_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    Dims,
    Encoding,
    Hash,
    ImageData,
    Schema,
    Timestamp,
)

import numpy as np

from . import schemas
from .utils import downsample_1d, downsample_2d, maybe_get


class FrameSelectionMode(enum.Enum):
    FRAME = "frame"
    CELL = "cell"
    PULSE = "pulse"


@dataclass
class PreviewSpec:
    # used for output schema
    name: str
    channel_name: Optional[str] = None  # set after init
    dimensions: int = 2
    wrap_in_imagedata: bool = True

    # reconfigurables in node schema
    swap_axes: bool = False  # only appars in schema if dimensions > 1
    flip_ss: bool = False  # only appars in schema if dimensions > 1
    flip_fs: bool = False
    nan_replacement: float = 0
    downsampling_factor: int = 1
    downsampling_function: Callable = np.nanmax
    # if frame_reduction, frame_reduction_fun is made automatically from index and mode
    # otherwise, leave None or provide custom
    frame_reduction: bool = True
    frame_reduction_selection_mode: FrameSelectionMode = FrameSelectionMode.FRAME
    # not all detectors provide cell and / or pulse IDs
    valid_frame_selection_modes: List[FrameSelectionMode] = tuple(FrameSelectionMode)
    frame_reduction_index: int = 0
    # either set explicitly or automatically with frame selection
    frame_reduction_fun: Optional[Callable] = None


class PreviewFriend:
    @staticmethod
    def add_schema(
        schema: Schema,
        outputs: List[PreviewSpec],
        node_name: str = "preview"
    ):
        """Add the prerequisite schema configurables for all the outputs."""
        (
            NODE_ELEMENT(schema)
            .key(node_name)
            .displayedName("Preview")
            .description(
                "Output specifically intended for preview in Karabo GUI. Includes "
                "some options for throttling and adjustments of the output data."
            )
            .commit(),
        )

        for spec in outputs:
            PreviewFriend._add_subschema(schema, f"{node_name}.{spec.name}", spec)

    @staticmethod
    def _add_subschema(schema, node_path, spec):
        # each preview gets its own node with config and channel
        (
            NODE_ELEMENT(schema)
            .key(node_path)
            .commit(),
        )

        if spec.dimensions > 1:
            (
                BOOL_ELEMENT(schema)
                .key(f"{node_path}.swapAxes")
                .displayedName("Swap slow / fast scan")
                .description(
                    "Swaps the two pixel axes around. Can be combined with "
                    "flipping to rotate the image"
                )
                .assignmentOptional()
                .defaultValue(spec.swap_axes)
                .reconfigurable()
                .commit(),

                BOOL_ELEMENT(schema)
                .key(f"{node_path}.flipSS")
                .displayedName("Flip SS")
                .description("Flip image data along slow scan axis.")
                .assignmentOptional()
                .defaultValue(spec.flip_ss)
                .reconfigurable()
                .commit(),
            )

        # configurable bits; these correspond to spec parts
        (
            BOOL_ELEMENT(schema)
            .key(f"{node_path}.flipFS")
            .displayedName("Flip FS")
            .description("Flip image data along fast scan axis.")
            .assignmentOptional()
            .defaultValue(spec.flip_fs)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key(f"{node_path}.downsamplingFactor")
            .displayedName("Downsampling factor")
            .description(
                "If greater than 1, the preview image will be downsampled by this "
                "factor before sending. This is mostly to save bandwidth in case GUI "
                "updates start lagging."
            )
            .assignmentOptional()
            .defaultValue(spec.downsampling_factor)
            .options("1,2,4,8")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{node_path}.downsamplingFunction")
            .displayedName("Downsampling function")
            .description("Reduction function used during downsampling.")
            .assignmentOptional()
            .defaultValue(spec.downsampling_function.__name__)
            .options("nanmax,nanmean,nanmin,nanmedian")
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{node_path}.replaceNanWith")
            .displayedName("NaN replacement")
            .description(
                "Displaying images in KaraboGUI seems to not go well when there are "
                "NaN values in data. And there will be with bad pixel masking or just "
                "geometry space between modules. NaN values get replaced with this "
                "value to get around this; choose a value which clearly stands out "
                "from the image data you want to see."
            )
            .assignmentOptional()
            .defaultValue(spec.nan_replacement)
            .reconfigurable()
            .commit(),
        )

        if spec.frame_reduction:
            (
                STRING_ELEMENT(schema)
                .key(f"{node_path}.selectionMode")
                .tags("managed")
                .displayedName("Index selection mode")
                .description(
                    "The value of preview.index can be used in multiple ways, "
                    "controlled by this value. If this is set to 'frame', "
                    "preview.index is sliced directly from data. If 'cell' "
                    "(or 'pulse') is selected, I will look at cell (or pulse) table "
                    "for the requested cell (or pulse ID). Special (stat) index values "
                    "<0 are not affected by this."
                )
                .options(
                    ",".join(
                        selectionmode.value
                        for selectionmode in spec.valid_frame_selection_modes
                    )
                )
                .assignmentOptional()
                .defaultValue("frame")
                .reconfigurable()
                .commit(),
                INT32_ELEMENT(schema)
                .key(f"{node_path}.index")
                .tags("managed")
                .displayedName("Index (or stat) for preview")
                .description(
                    "If this value is ≥ 0, the corresponding index (frame, cell, or "
                    "pulse) will be sliced for the preview output. If this value is "
                    "<0, preview will be one of the following stats: -1: max, "
                    "-2: mean, -3: sum, -4: stdev. These stats are computed across "
                    "frames, ignoring NaN values."
                )
                .assignmentOptional()
                .defaultValue(spec.frame_reduction_index)
                .minInc(-4)
                .reconfigurable()
                .commit(),
            )

        (
            OUTPUT_CHANNEL(schema)
            .key(f"{node_path}.output")
            .dataSchema(
                schemas.preview_schema(
                    wrap_image_in_imagedata=spec.wrap_in_imagedata
                )
            )
            .description("See description of grandparent node, 'preview'.")
            .commit(),
        )

    def __init__(self, device, outputs, node_name="preview"):
        self._device = device
        self.output_specs = outputs
        for spec in self.output_specs:
            spec.channel_name = f"{node_name}.{spec.name}.output"
        self.reconfigure(self._device.get(node_name))

    def write_outputs(
        self,
        *datas,
        timestamp=None,
        inplace=True,
        cell_table=None,
        pulse_table=None,
        warn_fun=None,
    ):
        """Applies GUI-friendly preview settings (replace NaN, downsample, wrap as
        ImageData) and writes to output channels. Make sure datas length matches number
        of channels! If inplace, masking and such may write to provided buffers (will
        otherwise copy)."""
        if isinstance(timestamp, Hash):
            timestamp = Timestamp.fromHashAttributes(
                timestamp.getAttributes("timestamp")
            )
        if warn_fun is None:
            warn_fun = self._device.log.WARN
        for data, spec in zip(datas, self.output_specs):
            if spec.frame_reduction_fun is not None:
                data = spec.frame_reduction_fun(data, cell_table, pulse_table, warn_fun)
            if spec.downsampling_factor > 1:
                if spec.dimensions == 1:
                    data = downsample_1d(
                        data,
                        spec.downsampling_factor,
                        reduction_fun=spec.downsampling_function,
                    )
                else:
                    data = downsample_2d(
                        data,
                        spec.downsampling_factor,
                        reduction_fun=spec.downsampling_function,
                    )
            elif not inplace:
                data = data.copy()
            if spec.flip_ss:
                data = np.flip(data, -2)
            if spec.flip_fs:
                data = np.flip(data, -1)
            if isinstance(data, np.ma.MaskedArray):
                data, mask = data.data, data.mask | ~np.isfinite(data)
            else:
                mask = ~np.isfinite(data)
            data[mask] = spec.nan_replacement
            mask = mask.astype(np.uint8)
            if spec.wrap_in_imagedata:
                output_hash = Hash(
                    "image.data",
                    ImageData(
                        maybe_get(data),
                        Dims(*data.shape),
                        Encoding.GRAY,
                        bitsPerPixel=32,
                    ),
                    "image.mask",
                    ImageData(
                        maybe_get(mask),
                        Dims(*mask.shape),
                        Encoding.GRAY,
                        bitsPerPixel=8,
                    ),
                )
            else:
                output_hash = Hash(
                    "image.data", maybe_get(data), "image.mask", maybe_get(mask)
                )
            self._device.writeChannel(
                spec.channel_name, output_hash, timestamp=timestamp, safeNDArray=True
            )

    def reconfigure(self, conf):
        for spec in self.output_specs:
            if not conf.has(spec.name):
                continue
            subconf = conf.get(spec.name)
            if spec.frame_reduction and (
                subconf.has("selectionMode") or subconf.has("index")
            ):
                if subconf.has("selectionMode"):
                    spec.frame_reduction_selection_mode = FrameSelectionMode(
                        subconf["selectionMode"]
                    )
                if subconf.has("index"):
                    spec.frame_reduction_index = subconf["index"]
                spec.frame_reduction_fun = _make_frame_reduction_fun(spec)
            if subconf.has("downsamplingFunction"):
                spec.downsampling_function = getattr(
                    np, subconf["downsamplingFunction"]
                )
            if subconf.has("downsamplingFactor"):
                spec.downsampling_factor = subconf["downsamplingFactor"]
            if subconf.has("flipSS"):
                spec.flip_ss = subconf["flipSS"]
            if subconf.has("flipFS"):
                spec.flip_fs = subconf["flipFS"]
            if subconf.has("swapAxes"):
                spec.flip_fs = subconf["flipFS"]
            if subconf.has("replaceNanWith"):
                spec.nan_replacement = subconf["replaceNanWith"]


def _make_frame_reduction_fun(spec) -> Optional[Callable]:
    """Will be called during configuration in case frame_reduction is True. Will
    use frame_reduction_selection_mode and frame_reduction_index to create a
    function to either slice or reduce input data."""
    # note: in case of frame picking, signature includes cell and pulse table
    if spec.frame_reduction_index < 0:
        # so while those tables are irrelevant here, still include in wrapper
        if spec.frame_reduction_index == -1:
            # note: separate from next case because dtype not applicable here
            reduction_fun = functools.partial(np.nanmax, axis=0)
        elif spec.frame_reduction_index == -2:
            # note: separate from next case because dtype not applicable here
            reduction_fun = functools.partial(np.nanmean, axis=0, dtype=np.float32)
        elif spec.frame_reduction_index == -3:
            # note: separate from next case because dtype not applicable here
            reduction_fun = functools.partial(np.nansum, axis=0, dtype=np.float32)
        elif spec.frame_reduction_index == -4:
            # note: separate from next case because dtype not applicable here
            reduction_fun = functools.partial(np.nanstd, axis=0, dtype=np.float32)

        def fun(data, cell_table, pulse_table, warn_fun):
            try:
                return reduction_fun(data)
            except Exception as ex:
                warn_fun(f"Frame reduction error: {ex}")

        return fun
    else:
        # here, we pick out a frame
        if spec.frame_reduction_selection_mode is FrameSelectionMode.FRAME:
            return _pick_preview_index_frame(spec.frame_reduction_index)
        elif spec.frame_reduction_selection_mode is FrameSelectionMode.CELL:
            return _pick_preview_index_cell(spec.frame_reduction_index)
        else:
            return _pick_preview_index_pulse(spec.frame_reduction_index)


def _pick_preview_index_frame(index):
    def aux(data, cell_table, pulse_table, warn_fun):
        if index >= data.shape[0]:
            warn_fun(
                f"Frame index {index} out of bounds for data with "
                f"{data.shape[0]} frames, previewing frame 0"
            )
            return data[0]
        return data[index]
    return aux


def _pick_preview_index_cell(index):
    def aux(data, cell_table, pulse_table, warn_fun):
        found = np.nonzero(cell_table == index)[0]
        if found.size == 0:
            warn_fun(
                f"Cell ID {index} not found in cell mapping, "
                "previewing frame 0 instead"
            )
            return data[0]
        if found.size > 1:
            warn_fun(
                f"Cell ID {index} not unique in cell mapping, "
                f"previewing first occurrence (frame {found[0]}"
            )
        return data[found[0]]
    return aux


def _pick_preview_index_pulse(index):
    def aux(data, cell_table, pulse_table, warn_fun):
        found = np.nonzero(pulse_table == index)[0]
        if found.size == 0:
            warn_fun(
                f"Pulse ID {index} not found in pulse table, "
                "previewing frame 0 instead"
            )
            return data[0]
        if found.size > 1:
            warn_fun(
                f"Pulse ID {index} not unique in pulse table, "
                f"previewing first occurrence (frame {found[0]})"
            )
        return data[found[0]]
    return aux
