import pytest

import numpy as np
from common_setup import OverridePythonDevice, dict_to_hash, kwhash
from karabo.bound import Configurator, PythonDevice

from calng.FrameSelectionArbiter import (
    SimpleFrameSelectionArbiter,
    AdvancedFrameSelectionArbiter,
)


@pytest.fixture
def advanced_arbiter():
    configurator = Configurator(PythonDevice)
    configurator.registerClass(AdvancedFrameSelectionArbiter)

    with OverridePythonDevice.override_context(PythonDevice):
        arbara = configurator.create(
            AdvancedFrameSelectionArbiter,
            kwhash(
                geometryDevice="geometry",
                sources=[],
                frameSelection=kwhash(
                    decision="",
                ),
            ),
        )
        arbara.initialization()
        yield arbara


@pytest.fixture
def simple_arbiter():
    configurator = Configurator(PythonDevice)
    configurator.registerClass(SimpleFrameSelectionArbiter)

    with OverridePythonDevice.override_context(PythonDevice):
        arbara = configurator.create(
            SimpleFrameSelectionArbiter,
            kwhash(
                geometryDevice="geometry",
                sources=[],
            ),
        )
        arbara.initialization()
        yield arbara


def test_simple(simple_arbiter):
    # don't care about matching, just use one source
    # only contains "cell table" to suggest number of frames
    sources_dict = {
        "arbitrarySource": (
            kwhash(
                image=kwhash(cellId=np.zeros(2)),
            ),
            None,
        ),
    }
    simple_arbiter.slotReconfigure(
        dict_to_hash({"frameSelection.kernelChoice": "ManualFilter"})
    )
    simple_arbiter.slotReconfigure(
        dict_to_hash(
            {"frameSelection.kernelParameters": {"spec": "0,1", "type": "mask"}}
        )
    )
    simple_arbiter.on_matched_data(1, sources_dict)
    assert simple_arbiter._last_written_data["data.dataFramePattern"] == [False, True]


def test_manual(simple_arbiter):
    # don't care about matching, just use one source
    # only contains "cell table" to suggest number of frames
    sources_dict = {
        "arbitrarySource": (
            kwhash(
                image=kwhash(cellId=np.zeros(100)),
            ),
            None,
        ),
    }
    simple_arbiter.slotReconfigure(
        dict_to_hash({"frameSelection.kernelChoice": "ManualFilter"})
    )

    include = [10, 20, 30]
    simple_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection.kernelParameters": {
                    "resize": "False",
                    "spec": ",".join(str(i) for i in include),
                    "type": "indices",
                }
            }
        )
    )
    simple_arbiter.on_matched_data(1, sources_dict)
    mask = simple_arbiter._last_written_data["data.dataFramePattern"]
    assert all(mask[i] for i in include)
    assert not any(mask[i] for i in range(len(mask)) if i not in include)

    include = list(range(1, 51, 3))
    simple_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection.kernelParameters": {
                    "resize": "False",
                    "spec": "1,51,3",
                    "type": "range",
                }
            }
        )
    )
    simple_arbiter.on_matched_data(2, sources_dict)
    mask = simple_arbiter._last_written_data["data.dataFramePattern"]
    assert all(mask[i] for i in include)
    assert not any(mask[i] for i in range(len(mask)) if i not in include)


def test_decision(advanced_arbiter):
    sources_dict = {
        "arbitrarySource": (
            kwhash(
                image=kwhash(cellId=np.zeros(2)),
            ),
            None,
        ),
    }

    advanced_arbiter.slotReconfigure(
        kwhash(
            frameSelection=kwhash(
                decision="a",
                plan=[
                    kwhash(
                        enable=True, name="a", kernel="ManualFilter", preselection=""
                    ),
                    kwhash(
                        enable=True, name="b", kernel="ManualFilter", preselection=""
                    ),
                ],
            )
        )
    )
    advanced_arbiter.slotReconfigure(
        dict_to_hash({"frameSelection.selections.a": {"type": "mask", "spec": "0,1"}}),
    )

    advanced_arbiter.on_matched_data(1, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [False, True]

    advanced_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection": {
                    "decision": "b",
                    "selections.b": {"type": "mask", "spec": "1,0"},
                }
            }
        )
    )

    advanced_arbiter.on_matched_data(2, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [True, False]

    advanced_arbiter.slotReconfigure(
        kwhash(
            frameSelection=kwhash(
                decision="a | b",
            )
        )
    )

    advanced_arbiter.on_matched_data(3, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [True, True]

    advanced_arbiter.slotReconfigure(
        kwhash(
            frameSelection=kwhash(
                decision="a & b",
            )
        )
    )

    advanced_arbiter.on_matched_data(4, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [
        False,
        False,
    ]


def test_preselection(advanced_arbiter):
    sources_dict = {
        "arbitrarySource": (
            dict_to_hash({"image.cellId": np.zeros(2)}),
            None,
        ),
    }
    advanced_arbiter.slotReconfigure(
        kwhash(
            frameSelection=kwhash(
                plan=[
                    kwhash(
                        enable=True, name="a", kernel="ManualFilter", preselection=""
                    ),
                    kwhash(
                        enable=True, name="b", kernel="ManualFilter", preselection="a"
                    ),
                ],
                decision="b",
            ),
        ),
    )
    # split in two calls as previous reconfigure should update schema
    advanced_arbiter.slotReconfigure(
        kwhash(
            frameSelection=kwhash(
                decision="b",
                selections=kwhash(
                    a=kwhash(spec="1,0"),
                    b=kwhash(spec="1,1"),
                ),
            )
        )
    )

    advanced_arbiter.on_matched_data(5, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [True, False]


def test_reduce_threshold(advanced_arbiter):
    sources_dict = {
        "sourceA": (
            kwhash(
                image=kwhash(cellId=np.zeros(3)),
                data=np.array([1, 1, -1], dtype=np.float32),
                data2=np.array([0, 0, -1], dtype=np.float32),
            ),
            None,
        ),
        "sourceB": (
            kwhash(
                data=np.array([0, 1, -1], dtype=np.float32),
                data2=np.array([1, 0, -1], dtype=np.float32),
            ),
            None,
        ),
    }
    advanced_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection": {
                    "plan": [
                        {  # just to test that preselection mask is applied
                            "enable": True,
                            "name": "a",
                            "kernel": "ManualFilter",
                            "preselection": "",
                        },
                        {
                            "enable": True,
                            "name": "b",
                            "kernel": "ReduceAndThreshold",
                            "preselection": "a",
                        },
                    ],
                    "decision": "b",
                }
            }
        )
    )

    advanced_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection.selections.a.spec": "1,1,0",
                "frameSelection.selections.b": {
                    "key": "data",
                    "threshold": 2,
                    "reduction": "sum",
                    "comparator": "lt",
                },
            }
        )
    )
    advanced_arbiter.on_matched_data(1, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [
        True,
        False,
        False,
    ]

    advanced_arbiter.slotReconfigure(
        dict_to_hash({"frameSelection.selections.b.key": "data2"})
    )
    advanced_arbiter.on_matched_data(2, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [
        True,
        True,
        False,
    ]

    advanced_arbiter.slotReconfigure(
        dict_to_hash({"frameSelection.selections.b.threshold": 1})
    )
    advanced_arbiter.on_matched_data(3, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [
        False,
        True,
        False,
    ]


def test_boolean_ops(advanced_arbiter):
    sources_dict = {
        "sourceA": (
            kwhash(
                image=kwhash(cellId=np.zeros(3)),
                data=np.array([1, 1, 1], dtype=bool),
                data2=np.array([0, 0, 1], dtype=bool),
            ),
            None,
        ),
        "sourceB": (
            kwhash(
                data=np.array([0, 1, 1], dtype=bool),
                data2=np.array([1, 0, 1], dtype=bool),
            ),
            None,
        ),
    }
    advanced_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection": {
                    "plan": [
                        {  # just to test that preselection mask is applied
                            "enable": True,
                            "name": "a",
                            "kernel": "ManualFilter",
                            "preselection": "",
                        },
                        {
                            "enable": True,
                            "name": "b",
                            "kernel": "BooleanCombination",
                            "preselection": "a",
                        },
                    ],
                    "decision": "b",
                }
            }
        )
    )

    advanced_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection.selections.a.spec": "1,1,0",
                "frameSelection.selections.b": {
                    "key": "data",
                    "operator": "all",
                },
            }
        )
    )
    advanced_arbiter.on_matched_data(1, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [
        False,
        True,
        False,
    ]

    advanced_arbiter.slotReconfigure(
        dict_to_hash({"frameSelection.selections.b.key": "data2"})
    )
    advanced_arbiter.on_matched_data(2, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [
        False,
        False,
        False,
    ]

    advanced_arbiter.slotReconfigure(
        dict_to_hash({"frameSelection.selections.b.operator": "any"})
    )
    advanced_arbiter.on_matched_data(3, sources_dict)
    assert advanced_arbiter._last_written_data["data.dataFramePattern"] == [
        True,
        False,
        False,
    ]


def test_random_frames(advanced_arbiter):
    np.random.seed(0)
    num_frames = 1000
    num_blocked = 100
    sources_dict = {
        "arbitrarySource": (
            kwhash(
                image=kwhash(cellId=np.zeros(num_frames)),
            ),
            None,
        ),
    }
    advanced_arbiter.slotReconfigure(
        dict_to_hash(
            {
                "frameSelection": {
                    "plan": [
                        {  # just to test that preselection mask is applied
                            "enable": True,
                            "name": "a",
                            "kernel": "ManualFilter",
                            "preselection": "",
                        },
                        {
                            "enable": True,
                            "name": "b",
                            "kernel": "RandomSampler",
                            "preselection": "a",
                        },
                    ],
                    "decision": "b",
                }
            }
        )
    )

    for prob in (10, 50, 90):
        advanced_arbiter.slotReconfigure(
            dict_to_hash(
                {
                    "frameSelection.selections.a": {
                        "spec": ",".join("0" * num_blocked),
                        "resize": "True",
                    },
                    "frameSelection.selections.b": {
                        "probability": prob,
                    },
                }
            )
        )
        advanced_arbiter.on_matched_data(1, sources_dict)
        res = advanced_arbiter._last_written_data["data.dataFramePattern"]
        assert not any(res[:num_blocked])
        expected = (num_frames - num_blocked) * (prob / 100)
        assert expected * 0.75 <= np.count_nonzero(res) <= expected * 1.25
